import enumerate.Action;
import enumerate.State;
import gameInterface.AIInterface;


public class Act {
	final Action myAction;
	final Action opAction;
	final int dist;
	final State myState;
	final State opState;
	Act lastMyAct = null;
	Act lastOpAct = null;
	int hitboxWidth;
	int energyCost;
	String myCharacter = AIInterface.CHARACTER_ZEN;
	String opCharacter = AIInterface.CHARACTER_ZEN;
	
	public Act(Action myAction, Action opAction, int dist, State myState, State opState, int hitboxWidth, int energyCost)
	{
		this.myAction = myAction;
		this.opAction = opAction;
		this.dist = dist;
		this.myState = myState;
		this.opState = opState;
		this.hitboxWidth = hitboxWidth;
		this.energyCost = energyCost;
		myCharacter = LittleFuzzy.MYCHAR;
		opCharacter = LittleFuzzy.OPCHAR;
	}
	
	public Act(CharInfoPack owner, CharInfoPack ownerOpponent, Act lastMyAct, Act lastOpAct, int hitboxWidth, int energyCost)
	{
		myAction = owner.action;
		opAction = ownerOpponent.action;
		dist = owner.box.distance(ownerOpponent.box);
		myState = owner.state;
		opState = ownerOpponent.state;
		this.hitboxWidth = hitboxWidth;
		this.energyCost = energyCost;

		this.lastMyAct = lastMyAct;
		this.lastOpAct = lastOpAct;
		myCharacter = LittleFuzzy.MYCHAR;
		opCharacter = LittleFuzzy.OPCHAR;
	}
	
	public Act(Act copy)
	{
		myAction = copy.myAction;
		opAction = copy.opAction;
		dist = copy.dist;
		myState = copy.myState;
		opState = copy.opState;

		lastMyAct = copy.lastMyAct;
		lastOpAct = copy.lastOpAct;
		hitboxWidth = copy.hitboxWidth;
		energyCost = copy.energyCost;
		myCharacter = copy.myCharacter;
		opCharacter = copy.opCharacter;
	}
	
	public Act(Act copy, Act lastMyAct, Act lastOpAct)
	{
		myAction = copy.myAction;
		opAction = copy.opAction;
		dist = copy.dist;
		myState = copy.myState;
		opState = copy.opState;
		hitboxWidth = copy.hitboxWidth;
		energyCost = copy.energyCost;
		myCharacter = copy.myCharacter;
		opCharacter = copy.opCharacter;

		this.lastMyAct = lastMyAct;
		this.lastOpAct = lastOpAct;
	}
	
	public double similarScenario(Act act)
	{
		double val = 0;
		if(opAction.name().equals(act.opAction.name())) //must be same action
			val+=0.5;
		if(myCharacter.toString().equals(act.myCharacter))
			val++;
		if(opCharacter.toString().equals(act.opCharacter))
			val+=0.5;
		val += 1 - Math.max(0, Math.abs(dist - act.dist) - act.hitboxWidth/4) / 10.0;
		/*if(myState != State.AIR || act.myState == State.AIR)
			val++;*/
		if(opState == act.opState)
			val++;
		
		if(lastMyAct == null || act.lastMyAct == null)
			val += 0.5;
		else
			val += 1 * lastMyAct.similarScenario(act.lastMyAct);
		if(lastOpAct == null || act.lastOpAct == null)
			val += 0.5;
		else
			val += 1 * lastOpAct.similarScenario(act.lastOpAct);
		
		val /= 6;
		return val;
	}
	
	public double similarAct(Act act)
	{
		if(!myAction.name().equals(act.myAction.name())) //must be same action
			return 0;
		return similarScenario(act);
		/*double val = 1;
		if(opAction.name().equals(act.opAction.name())) //must be same action
			val++;
		if(myCharacter.toString().equals(act.myCharacter) && opCharacter.toString().equals(act.opCharacter))
			val++;
		val += 1 - Math.max(0, Math.abs(dist - act.dist) - act.hitboxWidth/2) / 10.0;
		if(myState != State.AIR || act.myState == State.AIR)
			val++;
		if(opState == act.opState)
			val++;
		val /= 5;
		
		val *= 0.5;
		if(lastMyAct == null || act.lastMyAct == null)
			val += 0.125;
		else
			val += 0.25 * lastMyAct.similarAct(act.lastMyAct);
		if(lastOpAct == null || act.lastOpAct == null)
			val += 0.125;
		else
			val += 0.25 * lastOpAct.similarAct(act.lastOpAct);
		
		return val;*/
	}
	
	public static Act create(String s, int version)
	{
		try{
			if(s.equals("null"))
				return null;
			String[] args = s.split(":");
			int x = 0;
			if(isCharacter(args[1]))
				x = 2;
			Action myAction = Action.valueOf(args[x+1]);
			Action opAction = Action.valueOf(args[x+2]);
			int dist = Integer.valueOf(args[x+3]);
			int hitboxWidth = 0;
			if(version >= 2)
			{
				hitboxWidth = Integer.valueOf(args[x+4]);
				x++;
			}
			int energyCost = 0;
			if(version >= 3)
			{
				energyCost = Integer.valueOf(args[x+4]);
				x++;
			}
			State myState = State.valueOf(args[x+4]);
			State opState = State.valueOf(args[x+5]);
			Act lastMyAct = null;
			Act lastOpAct = null;
			if(!args[x+6].equals("<null>"))
				lastMyAct = Act.create(getSubAct(s,0),version);
			if(!args[x+7].equals("<null>"))
				lastOpAct = Act.create(getSubAct(s,1),version);
			
			Act cre = new Act(myAction, opAction, dist, myState, opState, hitboxWidth, energyCost);
			cre.lastMyAct = lastMyAct; cre.lastOpAct = lastOpAct;
			if(isCharacter(args[1]))
			{
				cre.myCharacter = args[x-1];
				cre.opCharacter = args[x];
			}
			
			return cre;
		} catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}
	
	public static boolean isCharacter(String s)
	{
		return s.equals(AIInterface.CHARACTER_GARNET) || s.equals(AIInterface.CHARACTER_KFM) 
				|| s.equals(AIInterface.CHARACTER_LUD) || s.equals(AIInterface.CHARACTER_ZEN);
	}
	
	protected static String getSubAct(String s, int index)
	{
		int open = 0, close = 0;
		int openIndex = 0;
		for(int i=s.indexOf('<'); i<s.length(); i++)
		{
			char c = s.charAt(i);
			if(c == '<')
			{
				if(open == 0)
					openIndex = i;
				open++;
			}
			else if(c == '>')
			{
				open--;
				if(open == 0)
				{
					if(close == index)
						return s.substring(openIndex+1, i);
					close++;
				}
			}
		}
		
		return s.substring(openIndex+1);
	}
	
	public String toString()
	{
		return "Act:"+myAction.name()+":"+opAction.name()+":"+dist+":"+hitboxWidth+":"+energyCost+":"+myState.name()+":"+opState.name()+":<"+lastMyAct+">:<"+lastOpAct+">";
		//return "Act:"+myCharacter+":"+opCharacter+":"+myAction.name()+":"+opAction.name()+":"+dist+":"+myState.name()+":"+opState.name()+":<"+lastMyAct+">:<"+lastOpAct+">";
	}

	
}
















