import enumerate.Action;
import enumerate.State;
import gameInterface.AIInterface;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.JFileChooser;
import javax.swing.JFrame;

import structs.CharacterData;
import structs.FrameData;
import structs.GameData;
import structs.Key;
import FuzzySystem.FuzzySystem;
import FuzzySystem.LinguisticList;
import FuzzySystem.MemberFunction;
import FuzzySystem.MemberGraph;
import FuzzySystem.RuleList;

import commandcenter.CommandCenter;


public class LittleFuzzy implements AIInterface {
	
	boolean sparringMode = false;
	boolean lessonMode = false;
	boolean cancelAttacks = false;
	Action lessonAction = null;
	int playerNumber;
	MemberGraph distance;
	LinguisticList input;
	LinguisticList output;
	FuzzySystem fuzzy;
	CommandCenter cc;
	LearningControl learn;
	
	static String MYCHAR;
	static String OPCHAR;
	
	/*public static void main(String[] args) {
		LittleFuzzy ai = new LittleFuzzy();
		ai.initialize(null,true);
		ai.input.get("distance").value = -90;
		ai.fuzzy.infer(ai.input, ai.output);
	}*/
	
	@Override
	public int initialize(GameData gameData, boolean playerNumber) {
		System.out.println("LittleFuzzy INITIALISING ======================");
		MYCHAR = gameData.getMyName(playerNumber);
		OPCHAR = gameData.getOpponentName(playerNumber);
		this.playerNumber = playerNumber? 1 : 2;
		fuzzy = new FuzzySystem(){
			@Override
			public RuleList create() {
				ArrayList<MemberFunction> funcs = new ArrayList<MemberFunction>();
				funcs.add(MemberFunction.leftShoulder("near", 0, 200));
				funcs.add(MemberFunction.rightShoulder("far", 80, 200));
				distance = new MemberGraph(funcs);
				input = new LinguisticList();
				input.graph(distance);
				input.add("distance");
				
				MemberGraph attack = MemberGraph.bool();
				MemberGraph move = MemberGraph.dir();
				output = new LinguisticList();
				output.graph(attack);
				output.add("attack");
				output.graph(move);
				output.add("move");
				RuleList movementBehaviour = new RuleList(distance, attack, move);
				movementBehaviour.addRule("if distance near then attack true, move backward");
				movementBehaviour.addRule("if distance far then attack false, move forward");
				
				return movementBehaviour;
			}
		};
		
		cc = new CommandCenter();
		learn = new LearningControl(this,2,3);
		loadInstructions();
		learn.knowledgeFile = MYCHAR+"_"+OPCHAR+".txt";
		learn.load();
		return 0;
	}
	
	public void loadInstructions()
	{
		String[] lines = read("instructions.txt");
		if(lines == null || lines.length == 0)
			return;
		
		for(String line : lines)
		{
			String[] args = line.split(" ");
			if(args[0].startsWith("sparring"))
				sparringMode = Boolean.valueOf(args[1]);
			else if(args[0].startsWith("lesson"))
				lessonMode = Boolean.valueOf(args[1]);
			else if(args[0].startsWith("cancelAttacks"))
				cancelAttacks = Boolean.valueOf(args[1]);
			else if(args[0].startsWith("knowledge"))
				learn.knowledgeFile = args[1];
		}
		
		if(sparringMode)
			System.out.println("Sparring Mode Active");
		if(lessonMode)
			System.out.println("Lesson Mode Active");
		System.out.println("Using knowledge from: "+learn.knowledgeFile);
	}
	
	public String[] read(String filename)
	{
		String lines = "";
		BufferedReader br = null;
		String line = "";
	 
		try {
			br = new BufferedReader(new FileReader("data/aiData/LittleFuzzy/"+filename));
			
			line = br.readLine();
			while(line != null)
			{
				lines += line+"\n";
			
				line = br.readLine();
			}
	 
		} catch (FileNotFoundException e) {
			//e.printStackTrace();
			return null;
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (br != null) {
				try {
					br.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		
		return lines.split("\n");
	}
	public static String[] read(File file)
	{
		String lines = "";
		BufferedReader br = null;
		String line = "";
	 
		try {
			br = new BufferedReader(new FileReader(file));
			
			line = br.readLine();
			while(line != null)
			{
				lines += line+"\n";
			
				line = br.readLine();
			}
	 
		} catch (FileNotFoundException e) {
			//e.printStackTrace();
			return null;
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (br != null) {
				try {
					br.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		
		return lines.split("\n");
	}
	
	public void write(String filename, String s)
	{
		PrintWriter file;
		try {
			file = new PrintWriter(new File("data/aiData/LittleFuzzy/"+filename));
			file.append(s);
			file.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	public static void write(File f, String s)
	{
		PrintWriter file;
		try {
			file = new PrintWriter(f);
			file.append(s);
			file.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	CharInfoPack me = null;
	CharInfoPack mePredicted = null;
	boolean meDirection;
	CharInfoPack opponent = null;
	CharInfoPack opPredicted = null;
	State predPrev;
	State oppPrev;
	Action opPrevAttack;
	Action myPrevAttack;
	int myhp = 0;
	int ophp = 0;
	int myenergy = 0;
	int openergy = 0;
	
	boolean first = true;
	
	//Save each frame, when hit, look at data 21 frames ago? Or look at last action and time taken
	
	boolean uniqueOp = true;
	boolean uniqueMy = true;
	
	private boolean uniqueOpAttack(String attack)
	{
		if(uniqueOp)
		{
			uniqueOp = false;
			return true;
		}
		if(opponent.remainingFrames == 1)
			uniqueOp = true;
		return false;
	}
	
	private boolean uniqueMyAttack(String attack)
	{
		if(uniqueMy)
		{
			uniqueMy = false;
			return true;
		}
		if(me.remainingFrames == 1)
			uniqueMy = true;
		return false;
	}
	
	@Override
	public void getInformation(FrameData frameData) {
		try{
			if(frameData == null)
				return;
			if(frameData.getP1() == null)
				return;
			if(frameData.getP2() == null)
				return;
			
			cc.setFrameData(frameData, playerNumber == 1);
			
			CharacterData data;
			if(playerNumber == 2)
				data = frameData.getP1();
			else
				data = frameData.getP2();
			HitBox box = new HitBox(data.getTop(), data.getLeft(), data.getBottom(), data.getRight());
			Speed speed = new Speed(data.getSpeedX(), data.getSpeedY());
			if(opponent != null)
			{
				oppPrev = opponent.state;
				opponent = new CharInfoPack(data.getState(),data.getAttack(),data.getAction(),box,speed,opponent.speed,
						data.getHp(),data.getEnergy(),data.getRemainingFrame());
			}
			else
				opponent = new CharInfoPack(data.getState(),data.getAttack(),data.getAction(),box,speed,null,
					data.getHp(),data.getEnergy(),data.getRemainingFrame());
			if(first)
			{
				CharInfoPack.neutral = opponent;
				first = false;
			}
			if(opPredicted != null)
				predPrev = opPredicted.state;
			opPredicted = opponent.prediction(20);
			
			if(playerNumber == 2)
				data = frameData.getP2();
			else
				data = frameData.getP1();
			
			meDirection = data.isFront();
			box = new HitBox(data.getTop(), data.getLeft(), data.getBottom(), data.getRight());
			speed = new Speed(data.getSpeedX(), data.getSpeedY());
			
			if(me != null)
				me = new CharInfoPack(data.getState(),data.getAttack(),data.getAction(),box,speed,me.speed,
						data.getHp(),data.getEnergy(),data.getRemainingFrame());
			else
				me = new CharInfoPack(data.getState(),data.getAttack(),data.getAction(),box,speed,null,
						data.getHp(),data.getEnergy(),data.getRemainingFrame());
			mePredicted = me.prediction(20);
			
		}catch(Exception e){
			e.printStackTrace();
		}
	}

	Random rnd = new Random();
	@Override
	public void processing() {
		if(first)
			return;
		
		try{
			if(cc.getskillFlag())
			{
				//k = cc.getSkillKey();
				//System.out.println("Skill Flag "+k.A+" "+k.B+" "+k.L+" "+k.R+" "+k.D+" "+k.U);
				return;
			}
			
			k.empty();
			if(cancelAttacks)
				cc.skillCancel();
			
			String op = opponent.action.name();
			String my = me.action.name();
			
			if(op.endsWith("A") || op.endsWith("B") || op.endsWith("C"))
			{
				if(uniqueOpAttack(op))
					learn.addOpAct();
			}
			else
				uniqueOp = true;
			if(my.endsWith("A") || my.endsWith("B") || my.endsWith("C"))
			{
				if(uniqueMyAttack(my))
					learn.addMyAct();
			}
			else
				uniqueMy = true;
			
			if(opponent.hp < ophp)
			{
				int damage = ophp - opponent.hp;
				ophp = opponent.hp;
				learn.whenHitting(damage);
			}
			if(me.hp < myhp)
			{
				int damage = myhp - me.hp;
				myhp = me.hp;
				learn.whenHit(damage);
			}
			
			myenergy = me.energy;
			openergy = opponent.energy;
			
			if(myPrevAttack != null && !myPrevAttack.equals(me.action))
			{
				String s = myPrevAttack.toString();
				if(isAttack(s))
				{
					System.out.println("my attack end: " + s);
					learn.whenIUse(myPrevAttack);
				}
			}
			if(opPrevAttack != null && !opPrevAttack.equals(opponent.action))
			{
				String s = opPrevAttack.toString();
				if(isAttack(s))
				{
					System.out.println("op attack end: " + s);
					learn.whenOpUses(opPrevAttack);
				}
			}
			
			myPrevAttack = me.action;
			opPrevAttack = opponent.action;
			
			/*if(opPredicted.state != predPrev || opponent.state != oppPrev)
				System.out.println(opponent.remainingFrames+" : "+opponent.box.bottom+" "+opPredicted.box.bottom +" : "+opponent.state+" "+opPredicted.state);*/
			
			int dist = opPredicted.box.distance(mePredicted.box);
			
			input.get("distance").value = dist;
			output.clear(0);
			fuzzy.infer(input,output);
			//System.out.println("move: "+output.get("move").value);
			//System.out.println("attack: "+output.get("attack").value);
			//System.out.println("dist: "+dist);
			if(output.get("move").value < -0.6)
			{
				if((meDirection && playerNumber == 1) || (meDirection && playerNumber == 2))
					k.L = true;
				else
					k.R = true;
			}
			else if(output.get("move").value > 0.3 && opponent.state != State.DOWN)
			{
				if((meDirection && playerNumber == 1) || (meDirection && playerNumber == 2))
					k.R = true;
				else
					k.L = true;
			}
			/*if(output.get("attack").value < 1)
				return;*/
			
			if(opPredicted.action == Action.DOWN)
				return;
			
			if(me.state != State.DOWN)
			{
				if(lessonMode)
				{
					if(lessonAction != null)
					{
						String com = actionToCall(lessonAction.name());
						System.out.println("Try: "+lessonAction.name()+" "+com);
						k.empty();
						cc.commandCall(com);
						return;
					}
				}
				else
				{
					RatedAct act = learn.getBestAct();
					if(act == null)
						return;
					if(!isAttack(opPredicted.action.name()) && Math.abs(dist - act.dist) > 10)
					{
						//System.out.println(dist+" "+act.dist);
						//return;
					}
					
					String com = actionToCall(act.myAction.name());
					//System.out.println("Try: "+act.myAction.name()+" "+com+" index: "+learn.lastChosenIndex);
					if(act.myAction.name().contains("AIR") && me.state != State.AIR)
						k.U = true;
					else
						cc.commandCall(com);
					return;
				}
			}
			
			/*if(meDirection && playerNumber == 1 || !meDirection && playerNumber == 2)
				k.L = true;
			else
				k.R = true;*/
			
			switch(opPredicted.state)
			{
			case AIR:
				cc.commandCall(actionToCall(Action.AIR_GUARD.name()));
				System.out.println("Guard Air");
				break;
			case CROUCH:
				cc.commandCall(actionToCall(Action.CROUCH_GUARD.name()));
				System.out.println("Guard Crouch");
				break;
			default:
				//cc.commandCall(actionToCall(Action.CROUCH_GUARD.name()));
				cc.commandCall(actionToCall(Action.STAND_GUARD.name()));
				System.out.println("Guard Stand");
				break;
			}
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	public boolean isAttack(String s)
	{
		return s.endsWith("A") || s.endsWith("B") || s.endsWith("C");
	}

	Key k = new Key();
	@Override
	public Key input() {
		if(cc.getskillFlag())
		{
			k = cc.getSkillKey();
			//System.out.println("Skill Flag "+k.A+" "+k.B+" "+k.L+" "+k.R+" "+k.D+" "+k.U);
		}
		
		if(sparringMode)
		{
			k.A = (rnd.nextInt(10) > 4) ? true : false;
			k.B = (rnd.nextInt(10) > 4) ? true : false;
			k.C = (rnd.nextInt(10) > 4) ? true : false;
			k.U = (rnd.nextInt(10) > 4) ? true : false;
			k.D = (rnd.nextInt(10) > 4) ? true : false;
			k.L = (rnd.nextInt(10) > 4) ? true : false;
			k.R = (rnd.nextInt(10) > 4) ? true : false;
		}
		return k;
		
	}
	
	private String actionToCall(String command)
	{
		switch(command){
		case	"NEUTRAL"	:
			break;
		case	"STAND"	:
			break;
		case	"FORWARD_WALK"	:
			return ("6");
		case	"DASH"	:
			return ("6 5 6");
		case	"BACK_STEP"	:
			return ("4 5 4");
		case	"CROUCH"	:
			return ("2");
		case	"JUMP"	:
			return ("8");
		case	"FOR_JUMP"	:
			return ("9");
		case	"BACK_JUMP"	:
			return ("7");
		case	"AIR"	:
			break;
		case	"STAND_GUARD"	:
			return ("4");
		case	"CROUCH_GUARD"	:
			return ("1");
		case	"AIR_GUARD"	:
			return ("7");
		case	"STAND_GUARD_RECOV"	:
			break;
		case	"CROUCH_GUARD_RECOV"	:
			break;
		case	"AIR_GUARD_RECOV"	:
			break;
		case	"STAND_HIT"	:
			break;
		case	"CROUCH_HIT"	:
			break;
		case	"AIR_HIT"	:
			break;
		case	"CHANGE_DOWN"	:
			break;
		case	"DOWN"	:
			break;
		case	"RISE"	:
			break;
		case	"LANDING"	:
			break;
		case	"THROW_A"	:
			return ("4 _ A");
		case	"THROW_B"	:
			return ("4 _ B");
		case	"THROW_HIT"	:
			break;
		case	"THROW_SUFFER"	:
			break;
		case	"STAND_A"	:
			return ("A");
		case	"STAND_B"	:
			return ("B");
		case	"CROUCH_A"	:
			return ("2 _ A");
		case	"CROUCH_B"	:
			return ("2 _ B");
		case	"AIR_A"	:
			return ("8 _ A");
		case	"AIR_B"	:
			return ("8 _ B");
		case	"AIR_DA"	:
			return ("8 5 2 _ A");
		case	"AIR_DB"	:
			return ("8 5 2 _ B");
		case	"STAND_FA"	:
			return ("6 _ A");
		case	"STAND_FB"	:
			return ("6 _ B");
		case	"CROUCH_FA"	:
			return ("3 _ A");
		case	"CROUCH_FB"	:
			return ("3 _ B");
		case	"AIR_FA"	:
			return ("9 _ A");
		case	"AIR_FB"	:
			return ("9 _ B");
		case	"AIR_UA"	:
			return ("8 5 8 _ A");
		case	"AIR_UB"	:
			return ("8 5 8 _ B");
		case	"STAND_D_DF_FA"	:
			return ("2 3 6 _ A");
		case	"STAND_D_DF_FB"	:
			return ("2 3 6 _ B");
		case	"STAND_F_D_DFA"	:
			return ("6 2 3 _ A");
		case	"STAND_F_D_DFB"	:
			return ("6 2 3 _ B");
		case	"STAND_D_DB_BA"	:
			return ("2 1 4 _ A");
		case	"STAND_D_DB_BB"	:
			return ("2 1 4 _ B");
		case	"AIR_D_DF_FA"	:
			return ("8 2 3 6 _ A");
		case	"AIR_D_DF_FB"	:
			return ("8 5 2 3 6 _ B");
		case	"AIR_F_D_DFA"	:
			return ("8 5 6 2 3 _ A");
		case	"AIR_F_D_DFB"	:
			return ("8 5 6 2 3 _ B");
		case	"AIR_D_DB_BA"	:
			return ("8 5 2 1 4 _ A");
		case	"AIR_D_DB_BB"	:
			return ("8 5 2 1 4 _ B");
		case	"STAND_D_DF_FC"	:
			return ("2 3 6 _ C");
		}
		return command;
	
}

	public String getCharacter() {
		return CHARACTER_ZEN;
	}
	
	@Override
	public void close() {
		learn.save();
	}

	private static void compileGraphData(JFrame frame)
	{
		String allLines = "";
		JFileChooser fileChooser = new JFileChooser();
		fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		fileChooser.setAcceptAllFileFilterUsed(false);
		if(fileChooser.showOpenDialog(frame) == JFileChooser.APPROVE_OPTION) {
			File folder = fileChooser.getSelectedFile();
			File[] files = folder.listFiles();
			for(File f : files)
			{
				if(f.getName().contains("CompiledData"))
					continue;
				String[] lines = read(f);
				String[] name = f.getName().split("\\.")[0].split("_");
				//allLines += name[name.length-1]+",";
				for(String line : lines)
					allLines += name[name.length-1]+"," + line + "\n";
			}
			
			String toname = folder.getAbsolutePath()+"\\\\CompiledData.csv";
			write(new File(toname), allLines);
		}
		
	}
	
	/*public static void main(String[] args) {
		JFrame frame = new JFrame("Compile Graph");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		//add components
		Dimension dim = new Dimension(1000, 800);
		compileGraphData(null);

		frame.pack();
		frame.setLocationRelativeTo(null);
		frame.setVisible(true);
		frame.repaint();
	}*/
}
