package rule;

import enumerate.State;

/**
 * 
 * @author Kevin Majchrzak
 * 
 */
public class ShortAttackCounter extends Rule {

	public ShortAttackCounter(GameStateInterface gameState) {
		super(gameState);
	}

	@Override
	public boolean satisfied() {

		boolean isAttack = gameState.getEnemyMotionData().getAttackHitDamage() > 0;

		// only for attacks with no speed x and on ground and not for fireballs
		if ((!isAttack)
				|| (gameState.getEnemyMotionData().getSpeedX() != 0)
				|| gameState.getEnemyCharacterData().getAction().name()
						.startsWith("STAND_D_DF_F")
				|| (gameState.getEnemyCharacterData().getState() == State.AIR)) {
			return false;
		}

		boolean myCharacterOnLeftSide = gameState.getCommandCenter().getMyX() < gameState
				.getCommandCenter().getEnemyX();

		// check if my character is in enemy attack range
		int enemyAttackNearestHitBoxEdge; // attack hitbox
		int myNearestHitBoxEdge; // character hitbox

		if (myCharacterOnLeftSide) {
			// see Attack.materialise
			enemyAttackNearestHitBoxEdge = gameState.getEnemyCharacterData()
					.getX()
					+ 256
					- gameState.getEnemyMotionData().getAttackHit().getR();
			myNearestHitBoxEdge = gameState.getMyCharacterData().getRight();
		} else {
			enemyAttackNearestHitBoxEdge = gameState.getEnemyCharacterData()
					.getX()
					+ gameState.getEnemyMotionData().getAttackHit().getR();
			myNearestHitBoxEdge = gameState.getMyCharacterData().getLeft();
		}

		// distance between my character and enemy attack hitbox

		boolean willHit = true;

		if ((gameState.getEnemyCharacterData().isFront() && (myCharacterOnLeftSide))
				|| ((!gameState.getEnemyCharacterData().isFront()) && (!myCharacterOnLeftSide))) {
			// enemy is attacking in wrong direction
			willHit = false;
		} else {
			willHit = (myCharacterOnLeftSide && (myNearestHitBoxEdge >= enemyAttackNearestHitBoxEdge))
					|| ((!myCharacterOnLeftSide) && (myNearestHitBoxEdge <= enemyAttackNearestHitBoxEdge));
		}

		int delay = 15;
		int nowRemainingFrame = gameState.getEnemyCharacterData()
				.getRemainingFrame() - delay;

		return (!willHit)
				&& (gameState.getCommandCenter().getDistanceX() < 380)
				&& (nowRemainingFrame <= 20) && (nowRemainingFrame > 0)
				&& (ActiveComboRuleWrapper.getSingleton().rule == null);

	}

	@Override
	public void execute() {
		if (gameState.getMyCharacterData().getState() != State.AIR) {
			// counter from ground
			// highest priority for mega fireball (startup is very low anyway)
			if (gameState.getMyCharacterData().getEnergy() >= 300) {
				gameState.getCommandCenter().commandCall("STAND_D_DF_FC");
			} else {
				// counterslide
				gameState.getCommandCenter().commandCall("2 1 4 _ B");
			}

		} else {
			// counter from air
			// air slide will either hit enemy or dodge attack
			gameState.getCommandCenter().commandCall("AIR_D_DF_FB");
		}
	}

	@Override
	public int getPriority() {
		return -2;
	}

}
