package FuzzySystem;

import java.util.ArrayList;

public class MemberGraph {
	ArrayList<MemberFunction> members;
	
	public MemberGraph(ArrayList<MemberFunction> members)
	{
		this.members = members;
	}
	
	public static MemberGraph bool()
	{
		ArrayList<MemberFunction> members = new ArrayList<MemberFunction>();
		members.add(MemberFunction.singleton("false", 0));
		members.add(MemberFunction.singleton("true", 1));
		return new MemberGraph(members);
	}
	
	public static MemberGraph dir()
	{
		ArrayList<MemberFunction> members = new ArrayList<MemberFunction>();
		members.add(MemberFunction.singleton("forward", 1));
		members.add(MemberFunction.singleton("none", 0));
		members.add(MemberFunction.singleton("backward", -1));
		return new MemberGraph(members);
	}
	
	double getCentroid() //performs discretisation
	{
		double sumA = 0;
		double sumB = 0;

		double a = members.get(0).getA();
		double d = members.get(members.size()-1).getD();

		double step = (double)d/(MemberFunction.CENPRES-1);
		for(double x=a; x<=d+0.0000001; x+=step)
		{
			double y = getY(x);
			sumA += y*x;
			sumB += y;
		}

		return sumA / sumB;
	}
	
	double getY(double x)
	{
		double max = 0;
		for(int i=0; i<members.size(); i++)
		{
			double y = members.get(i).getY(x);
			if(y > max)
				max = y;
		}

		return max;
	}
	
	public MemberFunction get(String name)
	{
		for(MemberFunction func : members)
			if(func.name.equals(name))
				return func;
		return null;
	}
	
}
