package FuzzySystem;

import java.util.ArrayList;

public class RuleList {
	ArrayList<Rule> ruleList = new ArrayList<Rule>();
	
	ArrayList<MemberFunction> funcs = new ArrayList<MemberFunction>();
	LinguisticList reference;
	
	public RuleList(MemberGraph ... graphs)
	{
		for(int i=0; i<graphs.length; i++)
			funcs.addAll(graphs[i].members);
	}
	
	public RuleList addAll(LinguisticList list)
	{
		reference.addAll(list);
		return this;
	}
	
	public RuleList addRule(String ruleString)
	{
		Rule rule = new Rule();
		
		//if ANTECEDENT MEMFUNC LogicalOperator then CONSEQUENT MEMFUNC, CONSEQUENT MEMFUNC
		
		ruleString = ruleString.replace("if", "").trim();
		String[] args = ruleString.split("then");
		String[] condition = args[0].split(",");
		String[] effect = args[1].split(",");
		
		for(int i=0; i<condition.length; i++)
		{
			String[] params = condition[i].trim().split(" ");
			if(params.length > 2)
				rule.add(new Antecedent(params[0],get(params[1]),LogicalOperator.valueOf(params[2])));
			else
				rule.add(new Antecedent(params[0],get(params[1])));
		}
		
		for(int i=0; i<effect.length; i++)
		{
			String[] params = effect[i].trim().split(" ");
			rule.add(new Consequent(params[0],get(params[1])));
		}
		
		//TODO hierarchical effect
		
		ruleList.add(rule);
		return this;
	}
	
	private MemberFunction get(String name)
	{
		for(MemberFunction func : funcs)
			if(func.name.equalsIgnoreCase(name))
				return func;
		return null;
	}
	
	public void inference(LinguisticList inList, LinguisticList outList) //perform inference on rule list
	{
		double accumulatedFiringStrength = 0;
		
		for(int i=0; i<ruleList.size(); i++) //for each rule
		{
			Rule rule = ruleList.get(i);
			double firingStrength = rule.getRuleFiringStrength(inList); //get rule firing strength
			
			if(firingStrength > 0)
			{
				if(rule.effectBehaviour != null) //if rule has sub-behaviour
					rule.effectBehaviour.inference(inList, outList, firingStrength); //infer sub-behaviour
				else //if rule has consequents instead
				{
					accumulatedFiringStrength += firingStrength;
					for(Consequent con : rule.effect)
						con.applyEffect(outList, firingStrength);
				}
			}
		}
		
		if(accumulatedFiringStrength == 0) //if no rules fired or only sub-behaviours found
			return;
		
		for(LinguisticVariable outVar : outList.lingVars) //divide output by accumulated firing strength
			outVar.value /= accumulatedFiringStrength;
	}
	
	public void inference(LinguisticList inList, LinguisticList outList, double firingStrength)
	{
		LinguisticList outListSub = new LinguisticList();
		outListSub.graph(outList.temp);
		for(LinguisticVariable var : outList.lingVars) //sets blank output list for inference
			outListSub.add(var.name);

		inference(inList, outListSub);
		for(LinguisticVariable var : outListSub.lingVars)
			var.value *= firingStrength;

		for(int i=0; i<outList.lingVars.size(); i++) //adds inferred outputs to previous outputs
			outList.lingVars.get(i).value += outListSub.lingVars.get(i).value;
	}
	
	LinguisticList memory;
	public void memorize(LinguisticList inList)
	{
		memory = inList;
	}
}
