package FuzzySystem;

public class MemberFunction {
	String name;
	double a;
	double b;
	double c;
	double d;
	
	double centroid = Double.MIN_VALUE;
	
	static int CENPRES = 100;
	
	public MemberFunction(String name, double centroid)
	{
		this(name, centroid, centroid, centroid, centroid);
		this.centroid = centroid;
	}
	
	public MemberFunction(String name, double a, double b, double c, double d)
	{
		this.name = name;
		this.a = a;
		this.b = b;
		this.c = c;
		this.d = d;
	}
	
	public static MemberFunction singleton(String name, double centroid)
	{
		return new MemberFunction(name,centroid);
	}
	public static MemberFunction triangle(String name, double a, double bc, double d)
	{
		return new MemberFunction(name, a, bc, bc, d);
	}
	public static MemberFunction leftShoulder(String name, double c, double d)
	{
		return new MemberFunction(name, Integer.MIN_VALUE, Integer.MIN_VALUE, c, d);
	}
	public static MemberFunction rightShoulder(String name, double a, double b)
	{
		return new MemberFunction(name, a, b, Integer.MAX_VALUE, Integer.MAX_VALUE);
	}
	
	public double getA(){ return a; }
	public double getB(){ return b; }
	public double getC(){ return c; }
	public double getD(){ return d; }
	
	public double getY(double x)
	{
		if(x <= a) //left edge
			return 0;
		if(x >= d) //right edge
			return 0;
		if(x < b && b != a) //rising edge
			return (x-a)/(b-a);
		if(x > c && d != c) //falling edge
			return (d-x)/(d-c);
		return 1;
	}
	
	public double getX(double y, boolean risingEdge)
	{
		if(risingEdge)
		{
			if(y == 0)
				return a;
			if(y == 1)
				return b;
			return y*(b-a)+a;
		}
		else
		{
			if(y == 0)
				return d;
			if(y == 1)
				return c;
			return d-y*(d-c);
		}
	}
	
	public double getCentroid()
	{
		if(centroid != Double.MIN_VALUE)
			return centroid;

		if(b-a == d-c) //if symmetrical
			return b + (c-b)/2;

		//otherwise, use discretisation

		double sumA = 0;
		double sumB = 0;

		double step = (double)d/(CENPRES-1);
		for(double x=a; x<=d+0.0000001; x+=step) //handles double inprecision
		{
			double y = getY(x);
			sumA += y*x;
			sumB += y;
		}

		centroid = sumA / sumB;

		return centroid;
	}
}
